%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Bousi-Prolog interpreter

:- module(evaluator, [
		load_tpl/1,             % +File
		solve_goal/1,           % :Goal
		get_sim_equations/1,    % -Equations
		add_sim_equations/1     % +Equations
   ]).

:- use_module(directives).
:- use_module(flags).

% The following libraries aren't used by this module, but are imported
% in case they are needed by Bousi-Prolog programs or queries
:- use_module(library(help)).
:- use_module(library(lists)).
:- use_module(library(readutil)).
:- use_module(library(statistics)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Public predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% load_tpl(+File)
%
%     Loads a translated Bousi-Prolog (TPL) file directly into memory,
%     unloading the last file loaded with this predicate first. If File
%     is an empty string, the current loaded file will be removed from
%     the database but no file will be loaded afterwards.
%

load_tpl(File) :-
	% Removes all dynamic predicates from the previous loaded program
	retractall(sim(_, _, _)),
	retractall(lEqThan(_, _, _)),
	retractall(gEqThan(_, _, _)),
	retractall(frel1(_, _, _)),
	retractall(frel2(_, _, _)),
	retractall(frel3(_, _, _)),
	(current_predicate(unload_file/1) ->
		% The unload_file/1 predicate is available only in SWI-Prolog
		% 5.10.0 and later
		current_file(OldFile),
		(OldFile \== '' ->
			unload_file(OldFile)
		;
			true
		),
		retract(current_file(OldFile)),
		assert(current_file(File))
	;
		% User-defined predicates can't be erased using retractall/1
		% because they're not dynamic; as a result, SWI-Prolog will
		% show a warning if the same program is loaded twice
		true
	),
	(File \== '' ->
		% Loads the new program
		consult(File)
	;
		% Initializes the symmetric fuzzy relations so they can be
		% used without having a program loaded
		assert(sim(X, X, 1.0)),
		assert(frel1(X, X, 1.0)),
		assert(frel2(X, X, 1.0)),
		assert(frel3(X, X, 1.0))
	).


%% solve_goal(:Goal)
%
%     Executes the specified Goal under this module, using the rules
%     and relations of the currently loaded program. Formally, this
%     predicate succeeds if there's a refutation for Goal with a
%     certain approximation degree.
%

solve_goal(Goal) :-
	catch((
		Goal
	% (catcher)
	), error(existence_error(procedure, Predicate), context(Context, Extra)), (
		% When an undefined predicate is called, SWI-Prolog throws an
		% exception that contains the internal name of some predicates
		% (e.g. given "undefined/2", the exception message will include
		% "evaluator:prefix_undefined/3"); here we get the actual names
		% of the predicates and throws a new exception with them		
		get_actual_predicate_definition(Predicate, ActualPredicate),
		get_actual_predicate_definition(Context, ActualContext),
		throw(error(existence_error(procedure, ActualPredicate), context(ActualContext, Extra)))
	)).


%% get_sim_equations(-Equations)
%
%     Unifies Equations with the list of currently defined
%     proximity/similarity equations, that is, sim/3 terms.
%

get_sim_equations(Equations) :-
	findall(sim(Sym1, Sym2, Degree), sim(Sym1, Sym2, Degree), Equations).


%% add_sim_equations(+Equations)
%
%     Asserts a list of proximity/similarity equations, that is, sim/3
%     terms. Equations that are already defined or aren't sim/3 terms
%     won't be added to the database.
%

add_sim_equations([]).

add_sim_equations([sim(Sym1, Sym2, Value)|MoreEquations]) :-
	sim(Sym1, Sym2, Value),
	% Equation is already defined, no need to define it again
	!,
	add_sim_equations(MoreEquations).

add_sim_equations([sim(Sym1, Sym2, Value)|MoreEquations]) :-
	% Equation isn't defined, so it must be asserted
	!,
	assert(sim(Sym1, Sym2, Value)),
	add_sim_equations(MoreEquations).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Weak unification algorithm
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% weak_unify(?Term1, ?Term2, +Lambda, ?Degree)
%
%     Unifies Term1 and Term2 using unification by proximity/similarity
%     and returns the approximation Degree in which both terms unify.
%     Lambda is the lower bound allowed for the approximation degree
%     of the weak unifications.
%

weak_unify(Atomic1, Atomic2, Lambda, Degree) :-
	% Atom (constant) unification
	atomic(Atomic1), atomic(Atomic2), !,
	sim(Atomic1, Atomic2, Degree),
	Degree >= Lambda.

weak_unify(Term1, Term2, Lambda, Degree) :-
	% Term decomposition
	compound(Term1), compound(Term2), !,
	Term1 =.. [Functor1|Args1],
	Term2 =.. [Functor2|Args2],
	length(Args1, Arity),
	length(Args2, Arity),
	sim(Functor1, Functor2, DegreeFunctor), 
	DegreeFunctor >= Lambda,
	weak_unify_args(Args1, Args2, Lambda, DegreeArgs),
	Degree is min(DegreeFunctor, DegreeArgs).

weak_unify(Term, Variable, _Lambda, 1) :-
	% Term/variable swap + Variable removal
	nonvar(Term), var(Variable), !,
	Variable = Term.

weak_unify(Variable, Term, _Lambda, 1) :-
	% Variable removal / Trivial equation removal
	var(Variable),
	Variable = Term.


%% weak_unify_args(?Args1, ?Args2, +Lambda, ?Degree)
%
%     Checks if the terms in the Args1 and Args2 lists can unify one
%     with each other and returns the minimum approximation Degree of
%     the unifications.
%

weak_unify_args([], [], _Lambda, 1).

weak_unify_args([Arg1|MoreArgs1], [Arg2|MoreArgs2], Lambda, Degree) :-
	weak_unify(Arg1, Arg2, Lambda, DegreeArg),
	weak_unify_args(MoreArgs1, MoreArgs2, Lambda, DegreeMoreArgs),
	Degree is min(DegreeArg, DegreeMoreArgs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Unification by proximity/similarity
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% unify(?Term1, ?Term2, ?Degree)
%
%     Unifies Term1 and Term2 using unification by proximity/similarity
%     and returns the approximation Degree in which both terms unify.
%

unify(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	weak_unify(Term1, Term2, Lambda, Degree).


%% unify(?Term1, ?Term2, +Comparer, ?Value)                  is semidet
%
%     Unifies Term1 and Term2 using unification by proximity/similarity
%     and succeeds only if the resulting approximation degree satisfies
%     the expression "degree Comparer Value" (for example, "degree >
%     0.5"). Comparer can be any Prolog arithmetic comparison operator
%     (=:=, =\=, >=, =<, >, <) or the unification operator (in the
%     latter case, Value will be unified with the approximation
%     degree).
%

unify(Term1, Term2, Comparer, Value) :-
	unify(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% unify_arguments(?Problems)
%
%     Solves several unification Problems using the weak unification
%     algorithm. Problems must be a list containing sublists with three
%     items: two terms and a variable where the approximation degree
%     of the terms will be stored. If any of the weak unifications
%     fail, the whole predicate will fail.
%

unify_arguments([]).

unify_arguments([[Term1, Term2, Degree]|MoreProblems]) :-
	unify(Term1, Term2, Degree),
	unify_arguments(MoreProblems).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Extension of user-defined fuzzy relations to terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% e_frel1(?Term1, ?Term2, ?Degree)                           is nondet
%
%     Compares Term1 and Term2 using '~1~' fuzzy relation and returns
%     the approximation Degree of the result (if it's greater than 0).
%

e_frel1(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	compare_terms(frel1, Term1, Term2, Lambda, Degree).


%% e_frel1(?Term1, ?Term2, +Comparer, ?Value)                is semidet
%
%     Compares Term1 and Term2 using '~1~' fuzzy relation and succeeds
%     only if the resulting approximation degree satisfies the
%     expression "degree Comparer Value" (for example, "degree > 0.5").
%     Comparer can be any Prolog arithmetic comparison operator (=:=,
%     =\=, >=, =<, >, <) or the unification operator (in the latter
%     case, Value will be unified with the approximation degree).
%

e_frel1(Term1, Term2, Comparer, Value) :-
	e_frel1(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% e_frel2(?Term1, ?Term2, ?Degree)                           is nondet
%
%     Compares Term1 and Term2 using '~2~' fuzzy relation and returns
%     the approximation Degree of the result (if it's greater than 0).
%

e_frel2(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	compare_terms(frel2, Term1, Term2, Lambda, Degree).


%% e_frel2(?Term1, ?Term2, +Comparer, ?Value)                is semidet
%
%     Compares Term1 and Term2 using '~2~' fuzzy relation and succeeds
%     only if the resulting approximation degree satisfies the
%     expression "degree Comparer Value" (for example, "degree > 0.5").
%     Comparer can be any Prolog arithmetic comparison operator (=:=,
%     =\=, >=, =<, >, <) or the unification operator (in the latter
%     case, Value will be unified with the approximation degree).
%

e_frel2(Term1, Term2, Comparer, Value) :-
	e_frel2(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% e_frel3(?Term1, ?Term2, ?Degree)                           is nondet
%
%     Compares Term1 and Term2 using '~3~' fuzzy relation and returns
%     the approximation Degree of the result (if it's greater than 0).
%

e_frel3(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	compare_terms(frel3, Term1, Term2, Lambda, Degree).


%% e_frel3(?Term1, ?Term2, +Comparer, ?Value)                is semidet
%
%     Compares Term1 and Term2 using '~3~' fuzzy relation and succeeds
%     only if the resulting approximation degree satisfies the
%     expression "degree Comparer Value" (for example, "degree > 0.5").
%     Comparer can be any Prolog arithmetic comparison operator (=:=,
%     =\=, >=, =<, >, <) or the unification operator (in the latter
%     case, Value will be unified with the approximation degree).
%

e_frel3(Term1, Term2, Comparer, Value) :-
	e_frel3(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% e_gEqThan(?Term1, ?Term2, ?Degree)                         is nondet
%
%     Compares Term1 and Term2 using '~>' ("less general than" /
%     "greater or equal than") fuzzy relation and returns the
%     approximation Degree of the result (if it's greater than 0).
%

e_gEqThan(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	compare_terms(gEqThan, Term1, Term2, Lambda, Degree).


%% e_gEqThan(?Term1, ?Term2, +Comparer, ?Value)              is semidet
%
%     Compares Term1 and Term2 using '~>' ("less general than" /
%     "greater or equal than") fuzzy relation and succeeds only if the
%     resulting approximation degree satisfies the expression
%     "degree Comparer Value" (for example, "degree > 0.5"). Comparer
%     can be any Prolog arithmetic comparison operator (=:=, =\=, >=,
%     =<, >, <) or the unification operator (in the latter case, Value
%     will be unified with the approximation degree).
%

e_gEqThan(Term1, Term2, Comparer, Value) :-
	e_gEqThan(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% e_lEqThan(?Term1, ?Term2, ?Degree)                         is nondet
%
%     Compares Term1 and Term2 using '<~' ("more general than" /
%     "less or equal than") fuzzy relation and returns the
%     approximation Degree of the result (if it's greater than 0).
%

e_lEqThan(Term1, Term2, Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	compare_terms(lEqThan, Term1, Term2, Lambda, Degree).


%% e_lEqThan(?Term1, ?Term2, +Comparer, ?Value)              is semidet
%
%     Compares Term1 and Term2 using '<~' ("more general than" /
%     "less or equal than") fuzzy relation and succeeds only if the
%     resulting approximation degree satisfies the expression
%     "degree Comparer Value" (for example, "degree > 0.5"). Comparer
%     can be any Prolog arithmetic comparison operator (=:=, =\=, >=,
%     =<, >, <) or the unification operator (in the latter case, Value
%     will be unified with the approximation degree).
%

e_lEqThan(Term1, Term2, Comparer, Value) :-
	e_lEqThan(Term1, Term2, Degree),
	apply(Comparer, [Degree, Value]),
	!.


%% compare_terms(+Relation, ?Term1, ?Term2, +Lambda, ?Degree)
%
%     Compares Term1 and Term2 using the specified fuzzy Relation and
%     Lambda-cut value, and returns the approximation Degree in which
%     both terms are similar.
%

compare_terms(Relation, Term1, Term2, _Lambda, Degree) :-
	% Variable-Variable comparison (two variables are related with
	% approximation degree 1 only if they're exactly the same variables
	% and the fuzzy relation is reflexive)
	var(Term1), var(Term2), !,
	Term1 == Term2,
	apply(Relation, [Term1, Term2, Degree]),
	var(Term1), var(Term2).

compare_terms(Relation, Term1, Term2, Lambda, Degree) :-
	% Atomic-Atomic comparison
	atomic(Term1), atomic(Term2), !,
	apply(Relation, [Term1, Term2, Degree]),
	Degree >= Lambda.

compare_terms(Relation, Term1, Term2, Lambda, Degree) :-
	% Compound-Compound comparison
	compound(Term1), compound(Term2), !,
	Term1 =.. [Functor1|Args1],
	Term2 =.. [Functor2|Args2],
	length(Args1, Arity),
	length(Args2, Arity),
	apply(Relation, [Functor1, Functor2, DegreeFunctor]),
	DegreeFunctor >= Lambda,
	compare_args(Args1, Args2, Relation, Lambda, DegreeArgs),
	Degree is min(DegreeFunctor, DegreeArgs).


%% compare_args(?Args1, ?Args2, +Relation, +Lambda, ?Degree)
%
%     Checks if the terms in the lists Args1 and Args2 are similar one
%     with each other using the given fuzzy Relation, and returns the
%     minimum approximation Degree of the results.
%

compare_args([], [], _Relation, _Lambda, 1).

compare_args([Arg1|MoreArgs1], [Arg2|MoreArgs2], Relation, Lambda, Degree) :-
	compare_terms(Relation, Arg1, Arg2, Lambda, DegreeArg),
	compare_args(MoreArgs1, MoreArgs2, Relation, Lambda, DegreeMoreArgs),
	Degree is min(DegreeArg, DegreeMoreArgs).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Helper predicates used by TPL programs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% eval_negation(+NegType, :Goal, +DegreeVars, -Degree)
%
%     Executes Goal and returns the approximation Degree of its
%     negation depending on the selected negation type ('not' or '\+').
%     DegreeVars is the list of degree variables that appears in Goal,
%     which is used to compute the approximation Degree of the result.
%

eval_negation(not, Goal, DegreeVars, Degree) :-
	% Weak Negation As Faliure
	(bpl_call(Goal), ! ->
		min_degree(DegreeVars, DegreeGoal),
		(DegreeGoal =:= 1 ->
			fail
		;
			Degree is 1 - DegreeGoal
		)
	;
		Degree is 1
	).

eval_negation(\+, Goal, DegreeVars, Degree) :-
	% Crisp Negation As Failure
	(bpl_call(Goal), ! ->
		min_degree(DegreeVars, DegreeGoal),
		(DegreeGoal =:= 1 ->
			fail
		;
			Degree is 1
		)
	;
		Degree is 1
	).


%% min_degree(+List, -MinDegree)                                 is det
%
%     Returns the smallest approximation degree found in List, ignoring
%     all items which aren't numbers. If List is an empty list or
%     doesn't contain any number, the maximum approximation degree (1)
%     is returned.
%

%% min_degree_aux(+List, -MinDegree)
%
%     Internal predicate used to avoid backtracking.
%
%     @see min_degree/2
%

min_degree(List, Minimum) :-
	min_degree_aux(List, Minimum), !.

min_degree_aux([], 1).

min_degree_aux([Number|List], Minimum) :-
	number(Number), !,
	min_degree_aux(List, PartialMinimum),
	Minimum is min(Number, PartialMinimum).

min_degree_aux([_NotANumber|List], Minimum) :-
	min_degree_aux(List, Minimum).


%% over_lambdacut(+Degree)                                   is semidet
%
%     Suceeds if the specified approximation Degree is greater than the
%     current lambda-cut value.
%

over_lambdacut(Degree) :-
	flags:get_bpl_flag(lambda_cut(Lambda)),
	Degree >= Lambda.


%% bpl_call(:Goal)
%
%     Extension of the call/1 predicate for Bousi-Prolog. This
%     predicate just executes specified Goal under this module.
%     However, if Goal is not a call to a Prolog predefined predicate
%     or a predicate currently loaded into this module, it'll be
%     appended a degree variable before executing it.
%

bpl_call(Goal) :-
	functor(Goal, Functor, Arity),
	current_predicate(Functor/Arity),
	% Just executes the goal
	!,
	Goal.

bpl_call(Goal) :-
	% Adds a new degree variable to the arguments of the goal
	Goal =.. [Functor|Args],
	append(Args, [_], NewArgs),
	% Appends the prefix of the currently loaded program to the functor
	flags:get_bpl_flag(program_prefix(ProgramPrefix)),
	concat_atom([ProgramPrefix, '_', Functor], NewFunctor),
	% Executes the new goal
	apply(NewFunctor, NewArgs).


%% bpl_apply(:Goal, +List)
%
%     Extension of the apply/2 predicate for Bousi-Prolog. This
%     predicate does exactly the same as apply/2, but executes the
%     resulting terms using bpl_call/1 instead of call/1.
%

bpl_apply(Term, List) :-
	Term =.. [Functor|Args],
	append(Args, List, NewArgs),
	NewTerm =.. [Functor|NewArgs],
	bpl_call(NewTerm).


%% bpl_maplist(:Goal, +List)
%
%     Extension of the maplist/2 predicate for Bousi-Prolog. This
%     predicate does exactly the same as maplist/2, but executes the
%     resulting terms using bpl_call/1 instead of call/1.
%

bpl_maplist(_Term, []).

bpl_maplist(Term, [Item|MoreItems]) :-
	Term =.. [Functor|Args],
	append(Args, [Item], NewArgs),
	NewTerm =.. [Functor|NewArgs],
	bpl_call(NewTerm),
	bpl_maplist(Term, MoreItems).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Miscellaneous predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% get_actual_predicate_definition(+InternalPredicate, -ActualPredicate)
%
%     Converts a TPL predicate definition into its original BPL predicate
%     definition, which doesn't include the "evaluator" module nor the
%     program prefix and has a smaller arity. If the predicate definition
%     isn't recognised as a TPL predicate, it'll be returned as is.
%

get_actual_predicate_definition(InternalPredicate, ActualPred/ActualArity) :-
	% Removes the "evaluator" module (if it's present), the program
	% prefix and decreases the arity; for example, if the currently
	% loaded program is "file.bpl", given "evaluator:file_procedure/3",
	% this predicate will return "procedure/2"
	(
		InternalPredicate = evaluator:Pred/Arity
	;
		InternalPredicate = Pred/Arity
	),
	flags:get_bpl_flag(program_prefix(Prefix)),
	atom_length(Prefix, PrefixLength),
	FullPrefixLength is PrefixLength + 1,
	sub_atom(Pred, 0, PrefixLength, _, Prefix),
	sub_atom(Pred, FullPrefixLength, _, 0, ActualPredAux),
	ActualArityAux is Arity - 1,
	(
		% Sometimes, undefined BPL predicates are translated twice; in
		% that cases, here we remove the program prefix and decrease
		% the arity again
		sub_atom(ActualPredAux, 0, PrefixLength, _, Prefix),
		sub_atom(ActualPredAux, FullPrefixLength, _, 0, ActualPred),
		ActualArity is ActualArityAux - 1
	;
		ActualPred = ActualPredAux,
		ActualArity = ActualArityAux
	),
	!.

get_actual_predicate_definition(evaluator:Pred/Arity, Pred/Arity) :-
	% Just removes the "evaluator" module from the predicate definition
	!.

get_actual_predicate_definition(Predicate, Predicate).
	% Returns the input predicate definition



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Dynamic predicates loaded from TPL files
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% sim(?Sym1, ?Sym2, ?Degree)                                is dynamic
%
%     Dynamic predicate which stores the proximity or similarity
%     equations defined in the current Bousi-Prolog program.
%

:- dynamic sim/3.


%% lEqThan(?Sym1, ?Sym2, ?Degree)                            is dynamic
%
%     Dynamic predicate which stores the "more general than" ("less or
%     equal than") equations defined in the current Bousi-Program
%     program.
%

:- dynamic lEqThan/3.


%% gEqThan(?Sym1, ?Sym2, ?Degree)                            is dynamic
%
%     Dynamic predicate which stores the "less general than" ("greater
%     or equal than") equations defined in the current Bousi-Program
%     program.
%

:- dynamic gEqThan/3.


%% frel1(?Sym1, ?Sym2, ?Degree)                              is dynamic
%
%     Dynamic predicate which stores the '~1~' fuzzy binary equations
%     defined in the current Bousi-Prolog program.
%

:- dynamic frel1/3.


%% frel2(?Sym1, ?Sym2, ?Degree)                              is dynamic
%
%     Dynamic predicate which stores the '~2~' fuzzy binary equations
%     defined in the current Bousi-Prolog program.
%

:- dynamic frel2/3.


%% frel3(?Sym1, ?Sym2, ?Degree)                              is dynamic
%
%     Dynamic predicate which stores the '~3~' fuzzy binary equations
%     defined in the current Bousi-Prolog program.
%

:- dynamic frel3/3.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Other dynamic predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% current_file(?File)                                       is dynamic
%
%     Dynamic predicate which contains the name of the last TPL file
%     loaded with load_tpl/1.
%
%     @see load_tpl/1
%

:- dynamic current_file/1.

current_file('').

